package org.bitbucket.bradleysmithllc.etlunit.feature.database;

import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql.ExecuteSqlHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql.ExecuteSqlRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_procedure.ExecuteSqlProcedureHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_procedure.ExecuteSqlProcedureRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_script.ExecuteSqlScriptHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_script.ExecuteSqlScriptRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.extend.Extender;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.File;
import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.Statement;
import java.util.List;

public class DbOperationsExecutor implements Extender, ExecuteSqlHandler, ExecuteSqlScriptHandler, ExecuteSqlProcedureHandler {
	private Log applicationLog;
	private final DatabaseFeatureModule databaseFeature;
	private DatabaseRuntimeSupport databaseRuntimeSupport;
	private RuntimeSupport runtimeSupport;
	private JDBCClient jdbcClient;

	public DbOperationsExecutor(DatabaseFeatureModule databaseFeature) {
		this.databaseFeature = databaseFeature;
	}

	@Inject
	public void setJdbcClient(JDBCClient client) {
		jdbcClient = client;
	}

	@Inject
	public void setDatabaseRuntimeSupport(DatabaseRuntimeSupport databaseRuntimeSupport) {
		this.databaseRuntimeSupport = databaseRuntimeSupport;
	}

	@Inject
	public void setRuntimeSupport(RuntimeSupport runtimeSupport) {
		this.runtimeSupport = runtimeSupport;
	}

	@Inject
	public void setLogger(@Named("applicationLog") Log log) {
		applicationLog = log;
	}

	public ClassResponder.action_code process(ETLTestMethod mt, ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject, VariableContext context, ExecutionContext executionContext) {
		return ClassResponder.action_code.handled;
	}

	public Feature getFeature() {
		return databaseFeature;
	}

	@Override
	public action_code executeSql(ExecuteSqlRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		String sqlText = request.getSql();

		applicationLog.info("executing statement " + sqlText);

		final String sql = sqlText;

		DatabaseClassListener.ConnectionMode connMode = DatabaseClassListener.getConnection(databaseFeature, valueObject, variableContext, testOperation.getTestMethod()).get(0);

		final DatabaseConnection conn = connMode.conn;

		final String mode = connMode.getMode();

		jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient() {
			public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
				SQLAggregator sqlagg = databaseFeature.resolveSQLRef(sql);

				SQLAggregator.Aggregator statementAggregator = sqlagg.getStatementAggregator();
				while (statementAggregator.hasNext()) {
					st.execute(statementAggregator.next().getLine());
				}
			}
		});

		return ClassResponder.action_code.handled;
	}

	@Override
	public action_code executeSqlProcedure(final ExecuteSqlProcedureRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		String script = request.getSqlProcedure().trim();

		final String argList;

		// dynamically create an argument list which exactly matches the supplied parameters
		StringBuilder argBuilder = new StringBuilder();

		List<String> orderedParameters = request.getParameters();

		if (orderedParameters != null && orderedParameters.size() != 0) {
			List pmap = (List) orderedParameters;

			for (int i = 0; i < pmap.size(); i++) {
				if (i != 0)
				{
					argBuilder.append(',');
				}

				argBuilder.append('?');
			}

			argList = " (" + argBuilder.toString() + ")";
		}
		else
		{
			argList = "()";
		}

		final String fscript = script;
		final String fscriptFull = fscript + argList;

		applicationLog.info("executing procedure " + fscript + " as " + fscriptFull);

		DatabaseClassListener.ConnectionMode connMode = DatabaseClassListener.getConnection(databaseFeature, valueObject, variableContext, testOperation.getTestMethod()).get(0);

		try {
			jdbcClient.useCallableStatement(connMode.conn, connMode.getMode(), new JDBCClient.CallableStatementClient() {
				@Override
				public String callText() {
					return "{call " + fscriptFull + "}";
				}

				@Override
				public void connection(Connection conn, CallableStatement st, DatabaseConnection connection, String mode, int id) throws Exception {
					List<String> orderedParameters = request.getParameters();

					if (orderedParameters != null && orderedParameters.size() != 0) {
						List pmap = (List) orderedParameters;

						for (int i = 0; i < pmap.size(); i++) {
							st.setString(i + 1, String.valueOf(pmap.get(i)));
						}
					}

					st.execute();
				}
			});
		} catch (TestExecutionError exe) {
			throw new TestExecutionError("", DatabaseConstants.ERR_SQL_SP_FAILURE, exe);
		}

		return action_code.handled;
	}

	@Override
	public action_code executeSqlScript(ExecuteSqlScriptRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		String script = request.getSqlScript();

		applicationLog.info("executing script " + script);

		File scriptFile = new FileBuilder(runtimeSupport.getCurrentTestResourceDirectory("sql")).mkdirs().name(script).file();

		try {
			final String sqlText = IOUtils.readFileToString(scriptFile);

			DatabaseClassListener.ConnectionMode connMode = DatabaseClassListener.getConnection(databaseFeature, valueObject, variableContext, testOperation.getTestMethod()).get(0);

			final DatabaseConnection conn = connMode.conn;

			final String mode = connMode.getMode();

			jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient() {
				public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
					SQLAggregator sqlagg = databaseFeature.resolveSQLRef(sqlText);

					SQLAggregator.Aggregator statementAggregator = sqlagg.getStatementAggregator();
					while (statementAggregator.hasNext()) {
						st.execute(statementAggregator.next().getLine());
					}
				}
			});

			return ClassResponder.action_code.handled;
		} catch (IOException e) {
			throw new TestExecutionError("", TestConstants.ERR_IO_ERROR, e);
		}
	}
}